﻿--[[
    Author: Valana <Team BATTLE> of Mug'thol Alliance-US
]]

--
-- Must not be local: relied upon by TbAutoBG:PARTY_MEMBERS_CHANGED() and TbAutoBG:OnEnable().
--
TbRaidHelper = LibStub("AceAddon-3.0"):NewAddon("TbRaidHelper", "AceConsole-3.0", "AceComm-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("TB_RaidHelper", true)
local TbCore = TbCore

local pairs =  pairs
local ipairs = ipairs
local tinsert = tinsert
local strfind = strfind
local format = format
local bband = bit.band
local wipe = wipe
local UninviteUnit = UninviteUnit
local IsRaidLeader = IsRaidLeader
local IsRaidOfficer = IsRaidOfficer
local GetNumRaidMembers = GetNumRaidMembers
local GetRaidRosterInfo = GetRaidRosterInfo
local GetNumLootItems = GetNumLootItems
local GetLootSlotInfo = GetLootSlotInfo
local GetLootSlotLink = GetLootSlotLink
local GetItemInfo = GetItemInfo
local LootSlot = LootSlot
local LootSlotIsCoin = LootSlotIsCoin
local UnitIsUnit = UnitIsUnit
local SendChatMessage = SendChatMessage
local UnitHealth = UnitHealth
local GetContainerNumSlots = GetContainerNumSlots

local _player = UnitName("player")
local _playerInRaid = bit.bor(COMBATLOG_OBJECT_TYPE_PLAYER, COMBATLOG_OBJECT_AFFILIATION_RAID)
local _raidRank
local _numAutoInviteKeywords = 0
local _pendingCalendarEventInvites = 0
local _vendored = { }
local _currency =
{
	["43016"] = true, -- Dalaran Cooking Award.
	["44990"] = true, -- Champion's Seal.
    ["29434"] = true,	-- Badge of Justice.
    ["40752"] = true,	-- Emblem of Heroism.
    ["40753"] = true,	-- Emblem of Valor.
    ["45624"] = true,	-- Emblem of Conquest.
    ["28558"] = true,	-- Spirit Shard.
    ["43228"] = true,	-- Stone Keeper's Shard.
}


-------------------------------------------------------------------------------
-- General help.
-------------------------------------------------------------------------------
TbRaidHelper.GeneralHelp = function()
    local _, name = GetAddOnInfo("TB_RaidHelper")
    TbCore:Print("---------------------")
    TbCore:Print(format(L["GENERAL_HELP_LINE1"], name .. " v" .. GetAddOnMetadata("TB_RaidHelper", "Version")))
    TbCore:Print("---------------------")
    TbCore:Print(L["GENERAL_HELP_LINE2"])
    TbCore:Print(L["GENERAL_HELP_LINE3"])
    TbCore:Print(L["GENERAL_HELP_LINE4"])
    TbCore:Print(L["GENERAL_HELP_LINE5"])
    TbCore:Print(L["GENERAL_HELP_LINE6"])
    TbCore:Print(L["GENERAL_HELP_LINE7"])
    TbCore:Print(L["GENERAL_HELP_LINE8"])
    TbCore:Print(L["GENERAL_HELP_LINE9"])
end


-------------------------------------------------------------------------------
-- Disband group.
-------------------------------------------------------------------------------
--
-- Slash command handler.
--
TbRaidHelper.DisbandGroup = function()
    --
    -- Do not disband inside an arena.
    --
    if (TbCore:IsInArena()) then return end

    if (TbCore:IsInBattleground()) then
        TbCore:PrintForChannel(L["DISBAND_IN_BG"], "SYSTEM", "SYSTEM")
    else
        local group = TbCore:GetRaidOrParty()
        if (not group) then
            TbCore:PrintForChannel(L["NOT_GROUPED"], "SYSTEM", "SYSTEM")
        elseif (IsRaidLeader() or IsRaidOfficer()) then
            SendChatMessage(format(L["DISBANDING"], L[group]), group)
            TbRaidHelper:ScheduleTimer(TbRaidHelper.DisbandHelper, .5, group)
        else
            TbCore:PrintForChannel(format(L["NOT_PROMOTED"], L[group]), "SYSTEM", "SYSTEM")
        end
    end
end

--
-- Actually disband the group.
--
TbRaidHelper.DisbandHelper = function(group)
    --
    -- Remove everyone. Members are numbered from 1, with the player being the highest-numbered member, followed by nil.
    --
    local num = 0
    if (group == "PARTY") then
        num = GetNumPartyMembers()
    elseif (group == "RAID") then
        num = GetNumRaidMembers() - 1
    end

    for index = 1, num do
        local current = group .. index
        if (TbCore:CanUninviteUnit(current)) then
            UninviteUnit(UnitName(current))
        end
    end

    --
    -- At the end, we might still be in the group with someone else (the raid leader). Leave the group to complete the process.
    --
    TbRaidHelper:ScheduleTimer(LeaveParty, .5)
end


-------------------------------------------------------------------------------
-- Order the recipient to drop group.
-------------------------------------------------------------------------------
TbRaidHelper.DropGroup = function(recipient)
    recipient = TbCore:StrTrim(recipient)
    if (TbCore:StrIsNilOrEmpty(recipient)) then
        LeaveParty()
        return
    end

    recipient = TbCore:Capitalize(UnitName(recipient) or recipient)
    TbCore:PrintForChannel(format(L["DROP_GROUP"], recipient), "SYSTEM", "SYSTEM")
    TbRaidHelper:SendCommMessage("TbRaidHelper:DROP_GROUP", "", "WHISPER", recipient)
end

function TbRaidHelper:DROP_GROUP(prefix, message, distribution, sender)
    --
    -- We don't want to drop our group while in an arena match or queued for a rated one, or in a PvE instance.
    --
    if (UnitIsUnit(_player, sender) or UnitIsDND("player") or TbCore:IsInArena() or TbCore:IsInRatedArenaQueue() or not TbCore:IsInGroup() or TbCore:IsInPvEInstance() or TbCore:IsIgnored(sender) or not TbCore:IsFriendOrGuildMate(sender)) then return end

    TbCore:PrintForChannel(format(L["LEAVING_GROUP"], sender), "SYSTEM", "SYSTEM")
    LeaveParty()
end


-------------------------------------------------------------------------------
-- Order the specified player to follow us.
-------------------------------------------------------------------------------
TbRaidHelper.Follow = function(follower)
    follower = TbCore:StrTrim(follower)
    if (TbCore:StrIsNilOrEmpty(follower)) then
        follower = "target"
    end
    if (TbCore:StrIsNilOrEmpty(follower) or UnitIsUnit(_player, follower)) then return end

    follower = TbCore:Capitalize(UnitName(follower) or follower)
    if (not CheckInteractDistance(follower, 4)) then
        TbCore:PrintForChannel(format(L["FOLLOW_TOO_FAR"], follower), "SYSTEM", "SYSTEM")
    else
        TbCore:PrintForChannel(format(L["FOLLOW"], follower), "SYSTEM", "SYSTEM")
        TbRaidHelper:SendCommMessage("TbRaidHelper:FOLLOW", "", "WHISPER", follower)
    end
end

function TbRaidHelper:FOLLOW(prefix, message, distribution, sender)
    if (UnitIsUnit(_player, sender) or UnitIsDND("player") or not CheckInteractDistance(sender, 4) or TbCore:IsIgnored(sender) or not TbCore:IsFriendOrGuildMate(sender)) then return end

    TbCore:PrintForChannel(format(L["FOLLOWING"], sender), "SYSTEM", "SYSTEM")
    FollowUnit(sender)
end


-------------------------------------------------------------------------------
-- Requests leadership or officership.
-------------------------------------------------------------------------------
TbRaidHelper.RequestLeadership = function(force)
    local token = "TbRaidHelper:PASS_LEADERSHIP"
    if (force) then token = "TbRaidHelper:FORCE_LEADERSHIP" end

    local group = TbCore:GetRaidOrParty()
    if (not group) then
        TbCore:PrintForChannel(L["NOT_GROUPED"], "SYSTEM", "SYSTEM")
    elseif (not force and (IsRaidLeader() or IsRaidOfficer())) then
        TbCore:PrintForChannel(L["ALREADY_PROMOTED"], "SYSTEM", "SYSTEM")
    else
        --
        -- Send an addon message requesting group leadership.
        --
        if (not force and group == "RAID") then
            TbCore:PrintForChannel(L["REQUESTING_OFFICERSHIP"], "SYSTEM", "SYSTEM")
        else
            TbCore:PrintForChannel(format(L["REQUESTING_LEADERSHIP"], L[group]), "SYSTEM", "SYSTEM")
        end
        TbRaidHelper:SendCommMessage(token, "", group)
    end
end

function TbRaidHelper:PASS_LEADERSHIP(prefix, message, distribution, sender)
    --
    -- Only react to requests from friends and guildmates as a safeguard.
    --
    if (UnitIsUnit("player", sender) or UnitIsDND("player") or not IsRaidLeader() or TbCore:IsInBattleground() or TbCore:IsIgnored(sender) or not TbCore:IsFriendOrGuildMate(sender)) then return end

    if (TbCore:IsInRaid() and strfind(prefix, "PASS_LEADERSHIP")) then
        TbCore:PrintForChannel(format(L["PROMOTING"], sender), "SYSTEM", "SYSTEM")
        PromoteToAssistant(sender)
    else
        TbCore:PrintForChannel(format(L["PASSING_LEADERSHIP"], sender), "SYSTEM", "SYSTEM")
        PromoteToLeader(sender)
    end
end


-------------------------------------------------------------------------------
-- Raid promotions and demotions.
-------------------------------------------------------------------------------
TbRaidHelper.PromoteAll = function()
    if (not TbCore:IsInRaid()) then
        TbCore:PrintForChannel(L["NOT_IN_RAID"], "SYSTEM", "SYSTEM")
    elseif (not IsRaidLeader()) then
        TbCore:PrintForChannel(L["NOT_RAID_LEADER"], "SYSTEM", "SYSTEM")
    else
        local PromoteToAssistant = PromoteToAssistant
        for index = 1, GetNumRaidMembers() - 1 do
            local name, rank = GetRaidRosterInfo(index)
            if (rank == 0 and not TbCore:IsIgnored(name)) then
                PromoteToAssistant("raid" .. index)
            end
        end
    end
end

TbRaidHelper.DemoteAll = function()
    if (not TbCore:IsInRaid()) then
        TbCore:PrintForChannel(L["NOT_IN_RAID"], "SYSTEM", "SYSTEM")
    elseif (not IsRaidLeader()) then
        TbCore:PrintForChannel(L["NOT_RAID_LEADER"], "SYSTEM", "SYSTEM")
    else
        local DemoteAssistant = DemoteAssistant
        for index = 1, GetNumRaidMembers() - 1 do
            local _, rank = GetRaidRosterInfo(index)
            if (rank == 1) then
                DemoteAssistant("raid" .. index)
            end
        end
    end
end

--
-- Notify the player when he or she gets promoted or demoted in a raid.
--
function TbRaidHelper:RAID_ROSTER_UPDATE(e)
    if (TbCore:IsInRaid()) then
        local _, raidRank = GetRaidRosterInfo(GetNumRaidMembers())
        if (raidRank ~= _raidRank) then
            if (raidRank == 2) then
                if (TbCore:IsInBattleground()) then
                    TbCore:PrintForChannel(ERR_NEW_LEADER_YOU, "SYSTEM", "SYSTEM")
                end
            elseif (raidRank == 1) then
                TbCore:PrintForChannel(L["PROMOTED"], "SYSTEM", "SYSTEM")
            elseif (raidRank == 0 and _raidRank == 1) then
                TbCore:PrintForChannel(L["DEMOTED"], "SYSTEM", "SYSTEM")
            end
            _raidRank = raidRank
        end
    else
        _raidRank = 0
    end
end


-------------------------------------------------------------------------------
-- Event Handlers.
-------------------------------------------------------------------------------
function TbRaidHelper:MERCHANT_SHOW(e)
    for bag = 0, 4 do
        for slot = 1, GetContainerNumSlots(bag) or 0 do
            --
            -- If we are no longer viewing the merchant window, abort.
            --
            if (not MerchantFrame:IsVisible() or MerchantFrame.selectedTab ~= 1) then break end

            --
            -- Sell grey-quality items.
            --
            local link = GetContainerItemLink(bag, slot)
            if (link) then
                local _, _, quality = GetItemInfo(link)
                if (quality == 0) then
                    tinsert(_vendored, link)
                    UseContainerItem(bag, slot)
                end
            end
        end
    end

    if (#_vendored > 0) then
        TbCore:PrintForChannel(format(L["AUTO_VENDORED"], TbCore:StrJoin(_vendored)), "SYSTEM", "SYSTEM")
        wipe(_vendored)
    end
end

function TbRaidHelper:LOOT_OPENED(e, isAutoLooting)
    --
    -- Loot the coins, unless we're auto-looting, to avoid the duplicate looting message.
    --
    local num = GetNumLootItems()
    local delay = 0
    if (isAutoLooting == 0 and LootSlotIsCoin(1)) then
        LootSlot(1)
        if (num < 2) then
            CloseLoot()
            return
        end
        delay = 0.5
    end

    --
    -- Automatically loot everything else, unless there is only one item and we are auto-looting (to avoid repeated looting sounds).
    --
    if (isAutoLooting == 0 or GetNumLootItems() > 1) then
		self:ScheduleTimer(self.LootRemainingItems, delay, num)
	end
end

TbRaidHelper.LootRemainingItems = function(num)
    if (num == 0) then
        if (GetNumLootItems() == 0) then
            CloseLoot()
        end
        return
    end

    local delay = .1
    if (LootSlotIsItem(num)) then
        local name, link, quality, _, _, category = GetItemInfo(GetLootSlotLink(num))
        if (quality < 2 and (category == "Quest" or category == "Key")) then
            --
            -- Loot keys quest and items, but prevent the index from being decremented so as to keep trying until they are confirmed as looted.
            --
            LootSlot(num)
            num = num + 1
            delay = 1
        else
			local id = strmatch(link, "\124Hitem:(%d-):.-\124h.-\124h")
			if (_currency[id]) then
				--
				-- Badges are now automatically distributed to all group members after one member clicks them, so try to loot them once then move on.
				--
				LootSlot(num)
				delay = 0.5
            end
        end
    end

    --
    -- Try again in 100 milliseconds if we are not looting anything, and in 1 second if we are.
    --
    TbRaidHelper:ScheduleTimer(TbRaidHelper.LootRemainingItems, delay, num - 1)
end


-------------------------------------------------------------------------------
-- Auto-invite and auto-accept.
-------------------------------------------------------------------------------
function TbRaidHelper:CHAT_MSG_WHISPER(e, msg, sender, ...)
    if (TbCore:IsIgnored(sender) or UnitIsUnit("player", sender) or UnitIsDND("player") or TbCore:IsInArena()) then return end

    --
    -- Check if the whispered message is an auto-invite keyword.
    --
    msg = TbCore:StrLowerTrim(msg)
    local matched = self.db.profile.autoInviteKeywords[msg]

    if (TbCore:IsInBattleground()) then
        if (matched) then
            --
            -- Warn the source that we might not be able to invite.
            --
            SendChatMessage(L["AUTO_INVITE_IN_BG"], "WHISPER", nil, sender)
            InviteUnit(sender)

            --
            -- Broadcast an auto-invite request to the battleground raid.
            --
            self:SendCommMessage("TbRaidHelper:AUTO_INVITE", sender, "RAID")
            return
        end
    elseif (not matched and TbCore:IsInActiveWintergrasp() and self:IsAutoInvitingInWintergrasp()) then
        for index, keyword in ipairs(L["WG_AUTO_INVITE_TRIGGERS"]) do
            if (strfind(msg, keyword)) then
                matched = true
                break
            end
        end
    end
    if (not matched) then return end

    if (TbCore:IsUnitInGroup(sender)) then
        SendChatMessage(L["ALREADY_IN_MY_GROUP"], "WHISPER", nil, sender)
    else
        local group = TbCore:GetRaidOrParty()
        if (not group) then
            InviteUnit(sender)
        elseif (IsRaidLeader() or IsRaidOfficer()) then
            if (TbCore:GroupHasRoom()) then
                InviteUnit(sender)
            else
                SendChatMessage(L["FULL_GROUP"], "WHISPER", nil, sender)
            end
        else
            --
            -- Warn the source that we do not have invite rights, and broadcast a request to the group.
            --
            SendChatMessage(L["CANNOT_INVITE"], "WHISPER", nil, sender)
            self:SendCommMessage("TbRaidHelper:AUTO_INVITE", sender, group)
        end
    end
end

--
-- Handle forwarded invite requests.
--
function TbRaidHelper:AUTO_INVITE(prefix, subject, distribution, sender)
    --
    -- We don't want to auto-invite people while in an arena match or queued for a rated one. Blindly attempt to invite while in a battleground.
    --
    if (UnitIsUnit("player", sender) or UnitIsDND("player") or (_numAutoInviteKeywords == 0 and (not TbCore:IsInActiveWintergrasp() or not self:IsAutoInvitingInWintergrasp())) or (not TbCore:IsInBattleground() and not TbCore:CanInvite()) or TbCore:IsInArena() or TbCore:IsInRatedArenaQueue() or TbCore:IsIgnored(sender) or TbCore:IsIgnored(subject)) then return end

    InviteUnit(subject)
end

--
-- Automatically accept party invitations from friends and guildmates.
--
function TbRaidHelper:PARTY_INVITE_REQUEST(e, source)
    if (TbCore:IsIgnored(source) or UnitIsDND("player") or not TbCore:IsFriendOrGuildMate(source)) then return end

    for i = 1, STATICPOPUP_NUMDIALOGS do
        local popup = _G["StaticPopup" .. i]
        if (popup and popup.which == "PARTY_INVITE") then
            StaticPopup_OnClick(popup, 1)
			TbCore:PrintRaidBossEmote(format(L["AUTO_ACCEPTED_INVITE"], source))
            return
        end
    end
end

--
-- Tell a player to drop group and whisper an auto-invite keyword if he or she was already grouped when we tried to invite him or her.
--
function TbRaidHelper:CHAT_MSG_SYSTEM(e, message)
    local recipient = strmatch(message, L["ALREADY_GROUPED"])
    if (not recipient) then return end

    if (TbCore:IsInActiveWintergrasp() and self:IsAutoInvitingInWintergrasp()) then
        SendChatMessage(format(L["DROP_GROUP_AND_WHISPER"], L["WG_AUTO_INVITE_SUGGEST"]), "WHISPER", nil, recipient)
    else
        --
        -- Pick the first general purpose auto-invite keyword.
        --
        for keyword, state in pairs(self.db.profile.autoInviteKeywords) do
            SendChatMessage(format(L["DROP_GROUP_AND_WHISPER"], keyword), "WHISPER", nil, recipient)
            break
        end
    end
end

TbRaidHelper.SystemMessageFilter = function(frame, e, message, ...)
    if (message == ERR_SET_LOOT_MASTER) then return true end
    if (not TbCore:IsInGroup() and strfind(message, L["DUNGEON_DIFFICULTY"])) then return true end
end

function TbRaidHelper:InviteUnit(unit, ...)
    if (not unit) then return end
    unit = TbCore:Capitalize(UnitName(unit) or unit)
    if (TbCore:IsUnitInGroup(unit)) then
        TbCore:PrintForChannel(format(L["ALREADY_IN_YOUR_GROUP"], unit), "SYSTEM", "SYSTEM")
    else
        return self.hooks["InviteUnit"](unit, ...)
    end
end


-------------------------------------------------------------------------------
-- Miscellaneous event handlers.
-------------------------------------------------------------------------------
function TbRaidHelper:UIErrorsFrame_AddMessage(frame, message, ...)
    --
    -- Suppress loot error messages resulting from auto-looting badges.
    --
    if (message ~= ERR_LOOT_CANT_LOOT_THAT_NOW) then return self.hooks[frame].AddMessage(frame, message, ...) end
end

--
-- Watch the combat log for the deaths of raid members not in our party.
--
function TbRaidHelper:COMBAT_LOG_EVENT_UNFILTERED(e, timeStamp, event, sourceGuid, sourceName, sourceFlags, destinationGuid, destinationName, destinationFlags, spellIdOrSwingDamage, spellNameOrSwingOverkill, spellSchool, spellDamage, spellOverkill, ...)
    --
    -- We are only interested in the deaths of friendly players in our raid, but not in our party. Announce them to all chat windows which
    -- display system messages. Note: we check the unit's health to avoid false positives with Feign Death.
    --
    if (not TbCore:IsInBattleground() and bband(destinationFlags, _playerInRaid) == _playerInRaid and bband(destinationFlags, COMBATLOG_OBJECT_AFFILIATION_PARTY) == 0 and event == "UNIT_DIED" and UnitHealth(destinationName) == 0 and not TbCore:IsIgnored(destinationName) and (not TbCore:IsInActiveWintergrasp() or TbCore:IsFriendOrGuildMate(destinationName))) then
        TbCore:PrintForChannel(format(ERR_PLAYER_DIED_S, destinationName), "SYSTEM", "SYSTEM")
    end
end


-------------------------------------------------------------------------------
-- Check calendar event invites.
-------------------------------------------------------------------------------
function TbRaidHelper:CheckPendingCalendarInvites(e)
    if (TbCore:IsInArena()) then return end

    --
    -- I think no CALENDAR_UPDATE_PENDING_INVITES is fired for past due calendar events which are moved forward, even if we never
    -- responded to them. For this reason, we initially subscribe to PLAYER_ENTERING_WORLD so the player will be warned about
    -- upon initial logon (or after zoning in the first time after a UI reload) if he or she does not open the calendar in-
    -- between.
    --
    self:UnregisterEvent("PLAYER_ENTERING_WORLD")

    --
    -- Only warn the player if the number of pending invites has increased since the last check.
    --
    local invites = CalendarGetNumPendingInvites()
    if (invites > _pendingCalendarEventInvites) then
        local message = L["PENDING_CALENDAR_INVITE"]
        if (invites > 1) then message = format(L["PENDING_CALENDAR_INVITES"], invites) end

        --
        -- Warn the player that there are pending calendar event invites waiting for his or her attention.
        --
        TbCore:PrintForChannel(message, "SYSTEM", "SYSTEM")
        TbCore:PrintRaidBossEmote(message, true)
    end
    _pendingCalendarEventInvites = invites
end

function TbRaidHelper:CALENDAR_EVENT_ALARM(e, title, hour, minute)
    TbCore:PrintRaidBossEmote(format(CALENDAR_EVENT_ALARM_MESSAGE, title), true)
end


-------------------------------------------------------------------------------
-- Initialization.
-------------------------------------------------------------------------------
function TbRaidHelper:OnInitialize()
    --
    -- Load our settings.
    --
    self.db = LibStub("AceDB-3.0"):New("TbRaidHelperDb",
    {
        profile =
        {
            autoInviteKeywords = { },
            autoInviteInWintergrasp = true,
        }
    }, "Default")
    self.db:SetProfile("Default")

--[===[@alpha@
    self.db.profile.autoInviteKeywords["tb"] = true
--@end-alpha@]===]

    --
    -- Create our options menu.
    --
    local _, addon = GetAddOnInfo("TB_RaidHelper")
    LibStub("AceConfig-3.0"):RegisterOptionsTable(addon,
        {
            type = "group",
            name = addon,
            desc = GetAddOnMetadata("TB_RaidHelper", L["DESCRIPTION_METADATA_FIELD"]),
            handler = TbRaidHelper,
            args =
            {
                subHeader =
                {
                    order = 0,
                    type = "description",
                    name = L["MENU_SUBHEADER"],
                },
                autoInviteHeader =
                {
                    order = 100,
                    type = "header",
                    name = L["MENU_AUTO_INVITE_HEADER"],
                },

                autoInviteSubHeader =
                {
                    order = 101,
                    type = "description",
                    name = L["MENU_AUTO_INVITE_SUBHEADER"],
                },
                wordList =
                {
                    order = 110,
                    type = "input",
                    name = L["MENU_AUTO_INVITE"],
                    desc = L["MENU_AUTO_INVITE_DESCRIPTION"],
                    usage = L["MENU_AUTO_INVITE_USAGE"],
                    width = "full",
                    multiline = true,
                    -- pattern = "[%w+[,;%s]]+",
                    get = "GetAutoInviteKeywords",
                    set = "SetAutoInviteKeywords",
                },
                inWintergrasp =
                {
                    order = 120,
                    type = "toggle",
                    name = L["MENU_IN_WINTERGRASP"],
                    desc = L["MENU_IN_WINTERGRASP_DESCRIPTION"],
                    get = "IsAutoInvitingInWintergrasp",
                    set = "ToggleAutoInviteInWintergrasp",
                },
            },
        })

    --
    -- Add our options menu to the Blizzard interface menu.
    --
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addon)

    --
    -- Define the general help slash command.
    --
    self:RegisterChatCommand(L["CMD_HELP1"], TbRaidHelper.GeneralHelp)
    self:RegisterChatCommand(L["CMD_HELP2"], TbRaidHelper.GeneralHelp)

    --
    -- Define the drop group slash command.
    --
    self:RegisterChatCommand(L["CMD_DROP1"], TbRaidHelper.DropGroup)

    --
    -- Define the follow me slash commands.
    --
    self:RegisterChatCommand(L["CMD_FOLLOW1"], TbRaidHelper.Follow)
    self:RegisterChatCommand(L["CMD_FOLLOW2"], TbRaidHelper.Follow)

    --
    -- Define the disband party/raid slash commands.
    --
    self:RegisterChatCommand(L["CMD_DISBAND1"], TbRaidHelper.DisbandGroup)
    self:RegisterChatCommand(L["CMD_DISBAND2"], TbRaidHelper.DisbandGroup)

    --
    -- Define the leadership request slash commands.
    --
    self:RegisterChatCommand(L["CMD_LEADERSHIP1"], function() TbRaidHelper.RequestLeadership() end)
    self:RegisterChatCommand(L["CMD_LEADERSHIP2"], function() TbRaidHelper.RequestLeadership(true) end)

    --
    -- Define the /promoteall command.
    --
    self:RegisterChatCommand(L["CMD_PROMOTE_ALL1"], TbRaidHelper.PromoteAll)
    self:RegisterChatCommand(L["CMD_PROMOTE_ALL2"], TbRaidHelper.PromoteAll)

    --
    -- Define the /demoteall command.
    --
    self:RegisterChatCommand(L["CMD_DEMOTE_ALL1"], TbRaidHelper.DemoteAll)
    self:RegisterChatCommand(L["CMD_DEMOTE_ALL2"], TbRaidHelper.DemoteAll)
end

function TbRaidHelper:OnEnable()
    --
    -- Apply all of our settings.
    --
    self:ApplyAllSettings()

    --
    -- Remove dungeon difficulty announcements while ungrouped.
    --
    ChatFrame_AddMessageEventFilter("CHAT_MSG_SYSTEM", TbRaidHelper.SystemMessageFilter)

    --
    -- Register for communications received from other players with TB Raid Helper.
    --
    self:RegisterComm("TbRaidHelper:PASS_LEADERSHIP", "PASS_LEADERSHIP")
    self:RegisterComm("TbRaidHelper:FORCE_LEADERSHIP", "PASS_LEADERSHIP")
    self:RegisterComm("TbRaidHelper:DROP_GROUP", "DROP_GROUP")
    self:RegisterComm("TbRaidHelper:FOLLOW", "FOLLOW")

    --
    -- Update event registrations.
    --
    self:RegisterEvent("PARTY_INVITE_REQUEST")
    self:RegisterEvent("RAID_ROSTER_UPDATE")
    self:RegisterEvent("MERCHANT_SHOW")
    self:RegisterEvent("LOOT_OPENED")
    self:RegisterEvent("CHAT_MSG_SYSTEM")
    self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    self:RegisterEvent("CALENDAR_EVENT_ALARM")
    self:RegisterEvent("CALENDAR_UPDATE_PENDING_INVITES", "CheckPendingCalendarInvites")
    self:RegisterEvent("PLAYER_ENTERING_WORLD", "CheckPendingCalendarInvites")

    --
    -- Hook the UI Errors frame to hide errors which may occur if the player auto-loots corpses independently of the addon.
    --
    self:RawHook(UIErrorsFrame, "AddMessage", "UIErrorsFrame_AddMessage", true)

    --
    -- Hook the InviteUnit function to prevent the player from inviting a group member.
    --
    self:RawHook("InviteUnit", "InviteUnit", true)
end

function TbRaidHelper:OnDisable()
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_SYSTEM", TbRaidHelper.SystemMessageFilter)
end


-------------------------------------------------------------------------------
-- Properties.
-------------------------------------------------------------------------------
function TbRaidHelper:GetAutoInviteKeywords()
    return TbCore:StrJoin(self.db.profile.autoInviteKeywords)
end

function TbRaidHelper:SetAutoInviteKeywords(info, list, ...)
    self.db.profile.autoInviteKeywords = TbCore:StrSplit(list, ",", true)
    self:ApplyAllSettings()
end

function TbRaidHelper:IsAutoInvitingInWintergrasp()
    return self.db.profile.autoInviteInWintergrasp
end

function TbRaidHelper:ToggleAutoInviteInWintergrasp()
    self.db.profile.autoInviteInWintergrasp = not self:IsAutoInvitingInWintergrasp()
    self:ApplyAllSettings()
end


-------------------------------------------------------------------------------
-- Apply the current settings.
-------------------------------------------------------------------------------
function TbRaidHelper:ApplyAllSettings()
    --
    -- Make sure to update _numAutoInviteKeywords here.
    --
    _numAutoInviteKeywords = TbCore:CountDictionaryEntries(self.db.profile.autoInviteKeywords)
    if (_numAutoInviteKeywords > 0 and self:IsAutoInvitingInWintergrasp()) then
        self:RegisterEvent("CHAT_MSG_WHISPER")
        self:RegisterComm("TbRaidHelper:AUTO_INVITE", "AUTO_INVITE")
    else
        self:UnregisterComm("TbRaidHelper:AUTO_INVITE")
        self:UnregisterEvent("CHAT_MSG_WHISPER")
    end
end
